from enum import Enum
from enum import IntEnum

class Anchor(Enum):
    LEFT_TOP = ("left" ,"top")
    LEFT_CENTER = ("left" ,"center")
    LEFT_BOTTOM = ("left" ,"bottom")
    CENTER_TOP = ("center" ,"top")
    CENTER = ("center" ,"center")
    CENTER_BOTTOM = ("center" ,"bottom")
    RIGHT_TOP = ("right" ,"top")
    RIGHT_CENTER = ("right" ,"center")
    RIGHT_BOTTOM = ("right" ,"bottom")

class MouseButton(IntEnum):
    NONE = 0
    LEFT = 1
    RIGHT = 2
    MIDDLE = 4
    X1 = 8
    X2 = 16

class KeyModifier(IntEnum):
    SHIFT = 1
    CTRL = 2
    ALT = 4
    META = 8

def checkbox_get(model, value):
    from .state import StateList, StateDict
    if isinstance(model.value, list) or isinstance(model.value, StateList):
        return value in model.value
    elif isinstance(model.value, dict) or isinstance(model.value, StateDict):
        return model.value.get(value, False)
    else:
        return bool(model.value)


def checkbox_set(model, checked, value):
    from .state import StateList, StateDict
    if isinstance(model.value, list) or isinstance(model.value, StateList):
        if checked:
            if not value in model.value:
                model.value.append(value)
        else:
            try:
                model.value.remove(value)
            except:
                pass
    elif isinstance(model.value, dict) or isinstance(model.value, StateDict):
        if checked:
            model.value[value] = True
        else:
            model.value.pop(value, None)
    else:
        model.value = checked



from .. import *
from .base import *

class Window(TkBaseWidget):
    pui_terminal = False
    def __init__(self, title=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.title = title
        self.size = size
        self.maximize = maximize
        self.fullscreen = fullscreen

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.curr_size = prev.curr_size
            self.curr_maximize = prev.curr_maximize
            self.curr_fullscreen = prev.curr_fullscreen
        else:
            self.ui = tk.Toplevel(self.parent.inner)
            # self.ui.protocol("WM_DELETE_WINDOW", self.on_closing)
            self.curr_size = Prop()
            self.curr_maximize = Prop()
            self.curr_fullscreen = Prop()

        if self.curr_size.set(self.size):
            self.ui.geometry("x".join([str(v) for v in self.size]))

        if self.curr_maximize.set(self.maximize):
            # https://stackoverflow.com/a/70061942
            try:
                self.ui.attributes('-zoomed', True)
            except:
                self.ui.state("zoomed")

        if self.curr_fullscreen.set(self.fullscreen):
            self.ui.attributes('-fullscreen', True)

        if not self.title is None:
            self.ui.title(self.title)

    def on_closing(self):
        self.get_node().tkparent.ui.destroy()

    def addChild(self, idx, child):
        if idx:
            return
        child.outer.pack(fill=tk.BOTH, expand=True)

    def removeChild(self, idx, child):
        child.outer.pack_forget()



from .. import *
from .base import *

class TextField(TBase):
    content_width = None
    def __init__(self, model, edit_model=None):
        super().__init__()
        self.model = model
        self.edit_model = edit_model
        self.editing = False

    def update(self, prev):
        model_value = str(self.model.value)

        if prev and prev.ui:
            self.editing = prev.editing
            self.ui = prev.ui
            self.curr_value = prev.curr_value
            self.changing = prev.changing

            if self.curr_value.set(model_value) and not self.editing:
                self.changing = True # block changed event from next line
                self.ui.value = model_value
        else:
            self.ui = widgets.Input(model_value)
            self.curr_value = Prop(model_value)
            self.changing = True # block changed event from initialization

        if self.edit_model and not self.editing:
            self.edit_model.value = model_value

        self.ui.puinode = self
        super().update(prev)

    def _tchanged(self, value): # editing
        if self.changing:
            self.changing = False
            return
        node = self.get_node()
        node.editing = True
        if node.edit_model:
            node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._input(e)

    def _tsubmitted(self, value): # finish editing
        node = self.get_node()
        node.editing = False
        node.model.value = value
        if node.edit_model:
            node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._change(e)


from .. import *
from .base import *

class TextField(WxBaseWidget):
    def __init__(self, model, edit_model=None):
        super().__init__()
        self.model = model
        self.edit_model = edit_model
        self.editing = False
        self.changed_cb = None
        self.layout_width = 50
        self.layout_height = -1

    def update(self, prev):
        model_value = str(self.model.value)
        if prev and prev.ui:
            self.editing = prev.editing
            self.ui = prev.ui
            self.curr_value = prev.curr_value
            if self.curr_value.set(model_value) and not self.editing:
                self.ui.SetValue(model_value)
        else:
            self.curr_value = Prop(model_value)
            self.ui = wx.TextCtrl(getWindow(self.parent), style=wx.TE_PROCESS_ENTER)
            self.ui.SetValue(model_value)
            self.ui.Bind(wx.EVT_TEXT, self.on_textchanged)
            self.ui.Bind(wx.EVT_KILL_FOCUS, self.on_kill_focus)
            self.ui.Bind(wx.EVT_TEXT_ENTER, self.on_enter)

        self.ui.SetMinSize((self.layout_width, self.layout_height))

        if self.edit_model and not self.editing:
            self.edit_model.value = model_value

        super().update(prev)

    def on_enter(self, event):
        from .window import Window
        p = self.parent
        while p:
            if isinstance(p, Window):
                p.panel.SetFocus()
                break
            p = p.parent
        event.Skip()

    def on_textchanged(self, event): # editing
        node = self.get_node()
        node.editing = True
        value = self.ui.GetValue()
        if node.edit_model:
           node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._input(e)

    def on_kill_focus(self, event): # finish editing
        node = self.get_node()
        node.editing = False
        value = self.ui.GetValue()
        node.model.value = value
        if node.edit_model:
            node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._change(e)



from .. import *
from .base import *
from .layout import *

class Divider(QtBaseWidget):
    pui_movable = False

    def __init__(self):
        super().__init__()

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QFrame()
            parent = self.non_virtual_parent
            if isinstance(parent, VBox):
                self.ui.setFrameShape(QtWidgets.QFrame.Shape.HLine)
            elif isinstance(parent, HBox):
                self.ui.setFrameShape(QtWidgets.QFrame.Shape.VLine)
            else:
                self.ui.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)
            self.ui.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        super().update(prev)



import threading
from .utils import *
from .dom import *

tls = threading.local()

class PuiViewNotFoundError(Exception): pass

def find_puiview():
    try:
        return tls.puistack[-1]
    except:
        raise PuiViewNotFoundError()

class PUIEvent():
    def __str__(self):
        return str(self.__dict__)


class PUINode():
    # To prevent crashes when we use a UI component not supported by the selected backend, this is useful when you are trying to support multiple backends.
    pui_supported = True

    # Tell DOM syncer not to go deeper into the node, it should be True for non-container elements
    pui_terminal = False

    # Menu and window-like UI elements, are out-of-order, so they are moved to the end of siblings before DOM syncing to simplify the process
    pui_outoforder = False

    # When enabled, children are added in reverse order, used by Stack
    pui_reversed_order = False

    # Used by PUIView
    pui_isview = False

    # Used by TimelineView, TabView, Sub-View or other nodes don't link to a real UI hierarchy
    # Children of virtual nodes will be promoted to the same level as the virtual node, addChild/removeChild won't be called for virtual nodes
    pui_virtual = False

    # When disabled, children are not allowed to be moved when DOM syncing, QSpacerItem is an example of non-movable node
    pui_movable = True

    # Used by grid layout, when enabled, sort children by grid_{row,column,rowspan,columnspan} before DOM sync
    pui_grid_layout = False

    def __init__(self, *args):
        from .view import PUIView

        if not hasattr(self, "name"):
            self.name = None

        self.destroyed = False
        self.retired_by = None
        self.pui_dom_parent = None
        self._debug = 0
        self._id = ""
        self._tag = ""

        self.layout_weight = None
        self.layout_width = None
        self.layout_height = None
        self.layout_padding = None
        self.layout_margin = None
        self.style_color = None
        self.style_bgcolor = None
        self.style_fontsize = None
        self.style_fontweight = None
        self.style_fontfamily = None
        self.grid_row = None
        self.grid_column = None
        self.grid_rowspan = None
        self.grid_columnspan = None

        self._onChanged = None
        self._onClicked = None
        self._onDblClicked = None
        self._onInput = None
        self._onKeyPress = None
        self._onMouseDown = None
        self._onMouseUp = None
        self._onMouseMove = None
        self._onWheel = None
        self._onDragEntered = None
        self._onDropped = None

        self.ui = None
        self.args = args
        try:
            self.root = find_puiview()
            self.parent = self.root.frames[-1]
        except PuiViewNotFoundError:
            self.root = self
            self.parent = self
            self.frames = []

        if isinstance(self, PUIView):
            self.root = self

        self.genKey()

        self.children = []

        if self.parent is self:
            self._path = tuple()
        else:
            self._path = self.parent._path + tuple([len(self.parent.children)])
            self.parent.children.append(self)

        # print(type(self).__name__, self._path, "parent=", self.parent._path)

    def findDomOffsetForNode(self, node):
        if self is node:
            return True, 0
        offset = 0
        for c in self.children:
            if c is node:
                return True, offset
            if c.pui_virtual:
                found, off = c.findDomOffsetForNode(node)
                offset += off
                if found:
                    return True, offset
            elif not c.pui_outoforder:
                offset += 1
        return False, offset

    def genKey(self):
        # key has to be relative to PUIView, so that it can be identical when a sub-PUIView is updated individually
        self.key = "|".join([x.name or type(x).__name__ for x in self.root.frames]+[self.name or type(self).__name__])
        if self.grid_row is not None and self.grid_column is not None:
            self.key += f":grid:{self.grid_row},{self.grid_column},{self.grid_rowspan},{self.grid_columnspan}"
        if hasattr(self, "_internal_tag"):
            self.key += f"%{self._internal_tag}"
        if self._id:
            self.key += f"#{self._id}"

    def __enter__(self):
        # print("enter", type(self).__name__, id(self))
        self.root.frames.append(self)
        return self

    def __exit__(self, ex_type, value, traceback):
        # print("exit", type(self).__name__, id(self))
        self.root.frames.pop()
        if ex_type is None: # don't consume exception
            return self

    @property
    def non_virtual_parent(self):
        p = self.parent
        while p.pui_virtual:
            p = p.parent
        return p

    @property
    def inner(self):
        if self.ui:
            return self.ui
        return self.parent.inner

    @property
    def outer(self):
        if self.ui:
            return self.ui
        if self.children:
            return self.children[0].outer
        return None

    def comment(self):
        return None

    def update(self, prev):
        if prev and prev is not self:
            prev.retired_by = self

    def postUpdate(self):
        pass

    def preSync(self):
        pass

    def postSync(self):
        pass

    def destroy(self, direct):
        self.root = None
        self.parent = None
        self.children = []

    def addChild(self, idx, child):
        pass

    def removeChild(self, idx, child):
        pass

    def debug(self, level=1):
        self._debug = level
        return self

    def id(self, name):
        self._id = name
        self.genKey()
        return self

    def tag(self, tag):
        self._tag = tag
        return self

    def get_node(self):
        node = self
        while node.retired_by:
            node = node.retired_by
        if node is not self:
            self.retired_by = node
        return node

    def __repr__(self):
        return self.serialize()

    def serialize(self, show_key=True, show_pyid=False, show_hierarchy=False, layout_debug=False):
        segs = []
        headline = [
            "  "*len(self._path),
            self.name or type(self).__name__,
        ]
        if show_pyid:
            headline.append(f"@{id(self)}")
        if self.children:
            headline.append(" {")

        # print view key
        if show_key:
            headline.append(" # Key:  ")
            headline.append(self.key)
            headline.append(" # Tag:  ")
            headline.append(self._tag)

        if show_hierarchy:
            headline.append(f" # parent={id(self.parent) if self.parent else None}")

        if layout_debug:
            headline.append(" # Layout:")
            if hasattr(self, "expand_x"):
                headline.append(f" expand_x={self.expand_x}")
            if hasattr(self, "expand_y"):
                headline.append(f" expand_y={self.expand_y}")
            if hasattr(self, "strong_expand_x"):
                headline.append(f" strong_expand_x={self.strong_expand_x}")
            if hasattr(self, "strong_expand_y"):
                headline.append(f" strong_expand_y={self.strong_expand_y}")
            if hasattr(self, "weak_expand_x"):
                headline.append(f" weak_expand_x={self.weak_expand_x}")
            if hasattr(self, "weak_expand_y"):
                headline.append(f" weak_expand_y={self.weak_expand_y}")
            if hasattr(self, "nweak_expand_x"):
                headline.append(f" nweak_expand_x={self.nweak_expand_x}")
            if hasattr(self, "nweak_expand_y"):
                headline.append(f" nweak_expand_y={self.nweak_expand_y}")
            if hasattr(self, "strong_expand_x_children"):
                headline.append(f" strong_expand_x_children={self.strong_expand_x_children}")
            if hasattr(self, "strong_expand_y_children"):
                headline.append(f" strong_expand_y_children={self.strong_expand_y_children}")

        if self.children:
            headline.append("\n")
        segs.append("".join(headline))

        comment = self.comment()
        if comment:
            segs.append("  "*(len(self._path)+1))
            segs.append("# ")
            segs.append(comment)
            segs.append("\n")

        if self.children:
            for i,c in enumerate(self.children):
                if i > 0:
                    segs.append("\n")
                segs.append(c.serialize(show_key=show_key, show_pyid=show_pyid, show_hierarchy=show_hierarchy, layout_debug=layout_debug))
            segs.append("\n")
            segs.append("".join(["  "*len(self._path), "}"]))
        return "".join(segs)

    def layout(self, width=None, height=None, weight=None, padding=None, margin=None):
        if not width is None:
            self.layout_width = width
        if not height is None:
            self.layout_height = height
        if not weight is None:
            self.layout_weight = weight
        if not padding is None:
            self.layout_padding = trbl(padding)
        if not margin is None:
            self.layout_margin = trbl(margin)

        return self

    def style(self, color=None, bgColor=None, fontSize=None, fontWeight=None, fontFamily=None):
        if not color is None:
            self.style_color = color
        if not bgColor is None:
            self.style_bgcolor = bgColor
        if not fontSize is None:
            self.style_fontsize = fontSize
        if not fontWeight is None:
            self.style_fontweight = fontWeight
        if not fontFamily is None:
            self.style_fontfamily = fontFamily

        return self

    def grid(self, row=None, column=None, rowspan=None, columnspan=None):
        if row is not None:
            self.grid_row = row
        if column is not None:
            self.grid_column = column
        if rowspan is not None:
            self.grid_rowspan = rowspan
        if columnspan is not None:
            self.grid_columnspan = columnspan
        self.genKey()
        return self

    def dragEnter(self, callback, *cb_args, **cb_kwargs):
        self._onDragEntered = callback, cb_args, cb_kwargs
        return self

    def drop(self, callback, *cb_args, **cb_kwargs):
        self._onDropped = callback, cb_args, cb_kwargs
        return self

    def click(self, callback, *cb_args, **cb_kwargs):
        self._onClicked = callback, cb_args, cb_kwargs
        return self

    def _clicked(self, e=None, *args, **kwargs):
        node = self.get_node()
        if node._onClicked:
            cb, cb_args, cb_kwargs = node._onClicked
            cb(e, *cb_args, **cb_kwargs)

    def dblclick(self, callback, *cb_args, **cb_kwargs):
        self._onDblClicked = callback, cb_args, cb_kwargs
        return self

    def _dblclicked(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onDblClicked:
            cb, cb_args, cb_kwargs = node._onDblClicked
            cb(e, *cb_args, **cb_kwargs)

    def change(self, callback, *cb_args, **cb_kwargs):
        self._onChanged = callback, cb_args, cb_kwargs
        return self

    def _change(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onChanged:
            cb, cb_args, cb_kwargs = node._onChanged
            cb(e, *cb_args, **cb_kwargs)

    def input(self, callback, *cb_args, **cb_kwargs):
        self._onInput = callback, cb_args, cb_kwargs
        return self

    def _input(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onInput:
            cb, cb_args, cb_kwargs = node._onInput
            cb(e, *cb_args, **cb_kwargs)

    def mousedown(self, callback, *cb_args, **cb_kwargs):
        self._onMouseDown = callback, cb_args, cb_kwargs
        return self

    def _mousedown(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onMouseDown:
            cb, cb_args, cb_kwargs = node._onMouseDown
            cb(e, *cb_args, **cb_kwargs)

    def mouseup(self, callback, *cb_args, **cb_kwargs):
        self._onMouseUp = callback, cb_args, cb_kwargs
        return self

    def _mouseup(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onMouseUp:
            cb, cb_args, cb_kwargs = node._onMouseUp
            cb(e, *cb_args, **cb_kwargs)

    def mousemove(self, callback, *cb_args, **cb_kwargs):
        self._onMouseMove = callback, cb_args, cb_kwargs
        return self

    def _mousemove(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onMouseMove:
            cb, cb_args, cb_kwargs = node._onMouseMove
            cb(e, *cb_args, **cb_kwargs)

    def wheel(self, callback, *cb_args, **cb_kwargs):
        self._onWheel = callback, cb_args, cb_kwargs
        return self

    def _wheel(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onWheel:
            cb, cb_args, cb_kwargs = node._onWheel
            cb(e, *cb_args, **cb_kwargs)

    def keypress(self, callback, *cb_args, **cb_kwargs):
        self._onKeyPress = callback, cb_args, cb_kwargs
        return self

    def _keypress(self, e, *args, **kwargs):
        node = self.get_node()
        if node._onKeyPress:
            cb, cb_args, cb_kwargs = node._onKeyPress
            cb(e, *cb_args, **cb_kwargs)

    def flet(self, **kwargs):
        return self

    def textual(self, **kwargs):
        return self

    def tkinter(self, **kwargs):
        return self

    def qt(self, **kwargs):
        return self



from .. import *
from .base import *

class Tabs(QtBaseWidget):
    NORTH = "n"
    SOUTH = "s"
    EAST = "e"
    WEST = "w"
    pui_terminal = False
    def __init__(self, tabposition=NORTH):
        super().__init__()
        self.tabposition = tabposition

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QTabWidget()
        super().update(prev)

    def addChild(self, idx, child):
        parent = child.parent
        while not isinstance(parent, Tab):
            parent = parent.parent
        self.ui.insertTab(idx, child.outer, parent.label)

    def removeChild(self, idx, child):
        self.ui.removeTab(idx)

    def postSync(self):
        for i,c in enumerate(self.children):
            self.ui.setTabText(i, c.label)


class Tab(PUINode):
    pui_virtual = True
    def __init__(self, label):
        super().__init__()
        self.label = label



from .. import *
from .base import *
import math

class Scroll(FBase):
    END = -0.0

    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        self.align_x = 0
        self.align_y = 0
        self.widget = None
        super().__init__()
        self.layout_weight = 1

    def update(self, prev):
        if prev and hasattr(prev, "hframe"):
            self.vframe = prev.vframe
            self.hframe = prev.hframe
        else:
            self.vframe = ft.Column() # outer
            self.hframe = ft.Row() # inner
            self.vframe.controls.append(self.hframe)
        self.vframe.expand = self.layout_weight
        if self.vertical is None:
            self.vframe.scroll = ft.ScrollMode.AUTO
        elif self.vertical:
            self.vframe.scroll = ft.ScrollMode.ADAPTIVE
        else:
            self.vframe.scroll = None
        if self.horizontal is None:
            self.hframe.scroll = ft.ScrollMode.AUTO
        elif self.horizontal:
            self.hframe.scroll = ft.ScrollMode.ADAPTIVE
        else:
            self.hframe.scroll = None
        try:
            self.vframe.update()
        except:
            pass
        super().update(prev)

    def addChild(self, idx, child):
        if idx != 0:
            return
        self.hframe.controls.insert(idx, child.outer)
        try:
            self.hframe.update()
        except:
            pass

    def removeChild(self, idx, child):
        if idx != 0:
            return
        self.hframe.controls.pop(idx)
        self.hframe.update()

    @property
    def outer(self):
        return self.vframe

    @property
    def inner(self):
        return self.hframe

    def scrollX(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_x = 0
            self.hsb_offset = pos
        else:
            self.align_x = 1
            self.hsb_offset = abs(pos)
        return self

    def scrollY(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_y = 0
            self.vsb_offset = pos
        else:
            self.align_y = 1
            self.vsb_offset = abs(pos)
        return self



from .view import *
from collections import defaultdict

class StateMutationInViewBuilderError(Exception):
    pass

class DummyBinding():
    def __init__(self, value):
        self.value = value

class BaseBinding():
    pass

class AttrBinding(BaseBinding):
    def __init__(self, state, key):
        try:
            self.viewroot = find_puiview()
            self.viewparent = self.viewroot.frames[-1]
        except:
            pass
        self.state = state
        self.key = key
        dt = type(getattr(self.state, self.key))
        if dt is str:
            self.func = str
        elif dt is int:
            self.func = int
        elif dt is float:
            self.func = float
        else:
            self.func = lambda x:x

    @property
    def value(self):
        return getattr(self.state, self.key)

    @value.setter
    def value(self, value):
        try: # skip validation error
            setattr(self.state, self.key, self.func(value))
        except:
            pass

    def change(self, callback):
        getattr(self.state, "_StateObject__callbacks")[self.key].add(callback)

    def bind(self, getter, setter):
        getattr(self.state, "_StateObject__binders")[self.key] = None
        setattr(getattr(self.state, "_StateObject__values"), self.key, getter())
        getattr(self.state, "_StateObject__binders")[self.key] = (getter, setter)

    def emit(self):
        listeners = set()
        for l in getattr(self.state, "_StateObject__listeners").values():
            listeners.update(l)
        _notify(getattr(self.state, "_StateObject__pending"), listeners)

class ListBinding(BaseBinding):
    def __init__(self, state, key):
        try:
            self.viewroot = find_puiview()
            self.viewparent = self.viewroot.frames[-1]
        except PuiViewNotFoundError:
            pass
        self.state = state
        self.key = key
        dt = type(self.state[self.key])
        if dt is str:
            self.func = str
        elif dt is int:
            self.func = int
        elif dt is float:
            self.func = float
        else:
            self.func = lambda x:x

    @property
    def value(self):
        try: # skip validation error
            return self.state[self.key]
        except:
            pass

    @value.setter
    def value(self, value):
        self.state[self.key] = self.func(value)

    def change(self, callback):
        getattr(self.state, "_StateList__callbacks")[self.key].add(callback)

    def bind(self, getter, setter):
        getattr(self.state, "_StateList__binders")[self.key] = None
        self.state[self.key] = getter()
        getattr(self.state, "_StateList__binders")[self.key] = (getter, setter)

    def emit(self):
        _notify(getattr(self.state, "_StateList__pending"), getattr(self.state, "_StateList__listeners"))

class DictBinding(BaseBinding):
    def __init__(self, state, key):
        try:
            self.viewroot = find_puiview()
            self.viewparent = self.viewroot.frames[-1]
        except PuiViewNotFoundError:
            pass
        self.state = state
        self.key = key
        dt = type(self.state[self.key])
        if dt is str:
            self.func = str
        elif dt is int:
            self.func = int
        elif dt is float:
            self.func = float
        else:
            self.func = lambda x:x

    @property
    def value(self):
        return self.state[self.key]

    @value.setter
    def value(self, value):
        try: # skip validation error
            self.state[self.key] = self.func(value)
        except:
            pass

    def change(self, callback):
        getattr(self.state, "_StateDict__callbacks")[self.key].add(callback)

    def bind(self, getter, setter):
        getattr(self.state, "_StateDict__binders")[self.key] = None
        self.state[self.key] = getter()
        getattr(self.state, "_StateDict__binders")[self.key] = (getter, setter)

    def emit(self):
        _notify(getattr(self.state, "_StateDict__pending"), getattr(self.state, "_StateDict__listeners"))

def _notify(pending, listeners):
    if pending is None:
        tbd = []
        for l in listeners:
            if l.retired_by:
                tbd.append(l)
        for l in tbd:
            listeners.remove(l)
        for l in list(listeners):
            l.redraw()
    else:
        pending.update(listeners)

class BaseState():
    pass

def Unstate(data):
    if isinstance(data, StateList):
        return [Unstate(item) for item in data._StateList__values]
    elif isinstance(data, StateDict):
        return {k: Unstate(v) for k, v in data._StateDict__values.items()}
    else:
        return data

def _state(data):
    if isinstance(data, list):
        return StateList([_state(item) for item in data])
    elif isinstance(data, dict):
        return StateDict({k: _state(v) for k, v in data.items()})
    else:
        return data

def State(data=None, deep=False):
    if data is None:
        return StateObject()
    if isinstance(data, list):
        if deep:
            return StateList([_state(item) for item in data])
        else:
            return StateList(data)
    if isinstance(data, dict):
        if deep:
            return StateDict({k: _state(v) for k, v in data.items()})
        else:
            return StateDict(data)
    return StateObject(data)

class StateObject(BaseState):
    def __init__(self, values=None):
        self.__listeners = defaultdict(set)
        self.__callbacks = defaultdict(set)
        self.__binders = {}
        self.__pending = None
        if values is None:
            self.__values = BaseState()
        else:
            self.__values = values

    def __call__(self, key=None):
        if key is None:
            listeners = set()
            for l in self.__listeners.values():
                listeners.update(l)
            _notify(self.__pending, listeners)
            return
        try:
            view = find_puiview()
            self.__listeners[key].add(view)
        except PuiViewNotFoundError:
            pass
        return AttrBinding(self, key)

    def __enter__(self):
        self.__pending = set()
        return self

    def __exit__(self, ex_type, value, traceback):
        pending = self.__pending
        self.__pending = None
        _notify(self.__pending, pending)

        if ex_type is None: # don't consume exception
            return self

    def __eq__(self, other):
        return self.__values == other

    # getter
    def __getattr__(self, key):
        if not key.startswith("_"):
            try:
                view = find_puiview()
                self.__listeners[key].add(view)
            except PuiViewNotFoundError:
                view = None
        ret = getattr(self.__values, key)
        if view:
            if isinstance(ret, StateObject):
                ret.__listeners[key].add(view)
            elif isinstance(ret, StateList):
                ret._StateList__listeners.add(view)
            elif isinstance(ret, StateDict):
                ret._StateDict__listeners.add(view)
        return ret

    # setter
    def __setattr__(self, key, value):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        if key.startswith("_"):
            object.__setattr__(self, key, value)
        else:
            if self.__binders.get(key):
                self.__binders[key][1](value)

            if isinstance(value, list):
                value = StateList(value)
            elif isinstance(value, dict):
                value = StateDict(value)
            if not hasattr(self.__values, key) or getattr(self.__values, key) != value:
                setattr(self.__values, key, value)
                _notify(self.__pending, self.__listeners[key])
                for cb in self.__callbacks[key]:
                    cb(value)

    # getter
    def __repr__(self):
        return f"StateObject({self.__values.__repr__()})"

class StateList(BaseState):
    def __init__(self, values=None):
        self.__listeners = set()
        self.__callbacks = defaultdict(set)
        self.__binders = {}
        self.__pending = None
        if values is None:
            self.__values = []
        else:
            self.__values = values

    def __call__(self, key=None):
        if key is None:
            _notify(self.__pending, self.__listeners)
            return
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return ListBinding(self, key)

    def __enter__(self):
        self.__pending = set()
        return self

    def __exit__(self, ex_type, value, traceback):
        pending = self.__pending
        self.__pending = None
        _notify(self.__pending, pending)

        if ex_type is None: # don't consume exception
            return self

    def __eq__(self, other):
        return self.__values == other

    # getter
    def __getitem__(self, key):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            view = None

        ret = self.__values[key]
        if view:
            if isinstance(ret, StateObject):
                ret._StateObject__listeners.add(view)
            elif isinstance(ret, StateList):
                ret.__listeners.add(view)
            elif isinstance(ret, StateDict):
                ret._StateDict__listeners.add(view)
        return ret

    # setter
    def __setitem__(self, key, value):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        if self.__binders.get(key):
            self.__binders[key][1](value)
        new = key >= len(self.__values)
        if new or self.__values[key] != value:
            self.__values[key] = value
            _notify(self.__pending, self.__listeners)
            for cb in self.__callbacks[key]:
                cb(value)
            if new:
                for cb in self.__callbacks[None]:
                    cb(value)

    # getter
    def __bool__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return bool(self.__values)

    # getter
    def __len__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return len(self.__values)

    # getter
    def __iter__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__iter__()

    # getter
    def __repr__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__repr__()

    # setter
    def append(self, obj):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.append(obj)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # setter
    def clear(self):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.clear()
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # getter
    def count(self, value):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.count(value)

    # setter
    def extend(self, iterable):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.extend(iterable)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # getter
    def index(self, value, *args, **kwargs):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.index(value, *args, **kwargs)

    # setter
    def insert(self, index, object):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.insert(index, object)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[index]:
            cb(object)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # getter/setter
    def pop(self, index=-1):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        r = self.__values.pop(index)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[index]:
            cb(self.__values)
        for cb in self.__callbacks[None]:
            cb(self.__values)
        return r

    # setter
    def remove(self, value):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.remove(value)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # setter
    def reverse(self, value):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.reverse(value)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)


    # setter
    def sort(self, *args, **kwargs):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.sort(*args, **kwargs)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # getter
    def get(self, index, default=None):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        if index >= 0 and index < len(self.__values):
            return self.__values[index]
        else:
            return default

    # getter
    def range(self):
        return range(len(self.__values))

class StateDict(BaseState):
    def __init__(self, values=None):
        self.__listeners = set()
        self.__callbacks = defaultdict(set)
        self.__binders = {}
        self.__pending = None
        if values is None:
            self.__values = {}
        else:
            self.__values = values

    def __call__(self, key=None):
        if key is None:
            _notify(self.__pending, self.__listeners)
            return
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return DictBinding(self, key)

    def __enter__(self):
        self.__pending = set()
        return self

    def __exit__(self, ex_type, value, traceback):
        pending = self.__pending
        self.__pending = None
        _notify(self.__pending, pending)

        if ex_type is None: # don't consume exception
            return self

    def __eq__(self, other):
        return self.__values == other

    # setter
    def __delitem__(self, key):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.__delitem__(key)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # getter
    def __getitem__(self, key):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            view = None
        ret = self.__values[key]
        if view:
            if isinstance(ret, StateObject):
                ret._StateObject__listeners.add(view)
            elif isinstance(ret, StateList):
                ret._StateList__listeners.add(view)
            elif isinstance(ret, StateDict):
                ret.__listeners.add(view)
        return ret

    # getter
    def __getattr__(self, key):
        if not key.startswith("_"):
            try:
                view = find_puiview()
                self.__listeners.add(view)
            except PuiViewNotFoundError:
                view = None
        ret = getattr(self.__values, key)
        if view:
            if isinstance(ret, StateObject):
                ret._StateObject__listeners.add(view)
            elif isinstance(ret, StateList):
                ret._StateList__listeners.add(view)
            elif isinstance(ret, StateDict):
                ret.__listeners.add(view)
        return ret

    # setter
    def __setattr__(self, key, value):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        if key.startswith("_"):
            object.__setattr__(self, key, value)
        else:
            if self.__binders.get(key):
                self.__binders[key][1](value)
            _notify(self.__pending, self.__listeners)
            return setattr(self.__values, key, value)

    # getter
    def __bool__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return bool(self.__values)

    # getter
    def __iter__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__iter__()

    # getter
    def __repr__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__repr__()

    # setter
    def __setitem__(self, key, value):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        if self.__binders.get(key):
            self.__binders[key][1](value)
        if not key in self.__values or self.__values[key] != value:
            self.__values[key] = value
            _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[key]:
            cb(value)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # setter
    def clear(self):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        self.__values.clear()
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    # getter
    def get(self, key, default=None):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.get(key, default)

    # getter
    def items(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.items()

    # getter
    def keys(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.keys()

    # getter/setter
    def pop(self, key, default=None):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        r = self.__values.pop(key, default)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)
        return r

    # setter
    def setdefault(self, key, default=None):
        try:
            find_puiview()
            raise StateMutationInViewBuilderError()
        except PuiViewNotFoundError:
            pass
        r = self.__values.setdefault(key, default)
        _notify(self.__pending, self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)
        return r

    # getter
    def values(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.values()



from .. import *
from .base import *

class ComboBox(QtBaseWidget):
    pui_terminal = False
    def __init__(self, editable=False, index_model=None, text_model=None):
        super().__init__()
        self.editable = editable
        self.index_model = index_model
        self.text_model = text_model # ComboBox can be editable so the value is processed as text

    def update(self, prev):
        if prev and prev.ui:
            self.curr_index = prev.curr_index
            self.curr_text = prev.curr_text
            self.signal_connected = prev.signal_connected
            self.ui = prev.ui
        else:
            self.curr_index = Prop()
            self.curr_text = Prop()
            self.signal_connected = False
            self.ui = QtWidgets.QComboBox()
        self.ui.setEditable(self.editable)
        super().update(prev)

    def postSync(self):
        index = -1
        text = ""

        if self.index_model:
            index = self.index_model.value
            text = self.children[index].text
        elif self.text_model:
            text = str(self.text_model.value)
            try:
                index = [str(c.value) for c in self.children].index(text)
            except:
                index = -1

        if self.signal_connected:
            self.ui.currentIndexChanged.disconnect()
            self.ui.currentTextChanged.disconnect()

        if self.curr_index.set(index):
            self.ui.setCurrentIndex(index)

        if self.curr_text.set(text):
            self.ui.setCurrentText(text)

        self.signal_connected = True
        self.ui.currentIndexChanged.connect(self.on_currentIndexChanged)
        self.ui.currentTextChanged.connect(self.on_currentTextChanged)


    def on_currentIndexChanged(self, idx):
        if self.index_model:
            self.index_model.value = idx
        if self.text_model:
            try:
                self.text_model.value = self.children[idx].value
            except:
                self.text_model.value = ""
        e = PUIEvent()
        e.value = idx
        self._change(e)

    def on_currentTextChanged(self, text):
        if not self.editable:
            return
        if self.text_model:
            self.text_model.value = text
        e = PUIEvent()
        e.value = text
        self._change(e)

    def addChild(self, idx, child):
        self.ui.insertItem(idx, child.text)

    def removeChild(self, idx, child):
        self.ui.removeItem(idx)

class ComboBoxItem(PUINode):
    def __init__(self, text, value=None):
        super().__init__()
        self.id(text)
        self.text = text
        if value is None:
            value = text
        self.value = value



class BaseTableAdapter():
    def data(self, row, col):
        """
        Return cell content
        """
        raise NotImplementedError("data() must be implemented")

    def editData(self, row, col):
        """
        Return data to be edited
        """
        return self.data(row, col)

    def setData(self, row, col, value):
        """
        Accepts the edited value
        """
        pass

    def editable(self, row, col):
        """
        Return whether the cell is editable
        """
        return False

    def columnHeader(self, col):
        """
        Return column header, set to None to hide column header
        """
        return f"Col {col}"

    def rowHeader(self, row):
        """
        Return row header, set to None to hide row header
        """
        return f"Row {row}"

    def rowCount(self):
        """
        Return number of rows
        """
        raise NotImplementedError("rowCount() must be implemented")

    def columnCount(self):
        """
        Return number of columns
        """
        raise NotImplementedError("columnCount() must be implemented")

class BaseTreeAdapter():
    def parent(self, node):
        """
        Return parent node
        """
        raise NotImplementedError("parent() must be implemented")

    def child(self, parent, index):
        """
        Return child node
        """
        raise NotImplementedError("child() must be implemented")

    def data(self, node):
        """
        Return node data
        """
        raise NotImplementedError("data() must be implemented")

    def rowCount(self, parent):
        """
        Return number of rows
        """
        raise NotImplementedError("rowCount() must be implemented")

    def clicked(self, node):
        """
        Called when a node is clicked
        """
        pass

    def dblclicked(self, node):
        """
        Called when a node is double clicked
        """
        pass

    def expanded(self, node):
        """
        Called when a node is expanded
        """
        pass

    def collapsed(self, node):
        """
        Called when a node is collapsed
        """
        pass



from .. import *
from .base import *

class Checkbox(TBase):
    def __init__(self, text, model, value=None):
        super().__init__()
        self.text = text
        self.model = model
        self.value = text if value is None else value

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.value = checkbox_get(self.model, self.value)
        else:
            self.ui = widgets.Checkbox(self.text, checkbox_get(self.model, self.value))

        self.ui.puinode = self
        super().update(prev)

    def _changed(self, value):
        checkbox_set(self.model, value, self.value)



from .application import *
from .button import *
from .checkbox import *
from .label import *
from .layout import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .tab import *
from .text import *
from .textfield import *
from .window import *
from .. import NotImplementedNode

PUIView = TPUIView

class DummyWidget(TBase):
    pui_supported = False
    def __init__(self, *args, **kwrgas):
        super().__init__()

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = widgets.Label("Not Implemented")



Canvas = DummyWidget
Combobox = DummyWidget
ComboboxItem = DummyWidget
Divider = lambda: None

PUI_BACKEND = "textual"



from .node import *
from .dom import *
from .state import StateMutationInViewBuilderError
import time

DEBUG = False

class PUIView(PUINode):
    pui_virtual = True
    pui_isview = True
    __ALLVIEWS__  = []

    @staticmethod
    def reload():
        for v in PUIView.__ALLVIEWS__:
            if DEBUG:
                print("reload", v.key)
            v.redraw()

    def __init__(self, *args):
        self.frames = []
        self.dirty = False
        self.updating = False
        super().__init__(*args)
        PUIView.__ALLVIEWS__.append(self)

    def __enter__(self):
        if not hasattr(tls, "puistack"):
            tls.puistack = []
        tls.puistack.append(self)
        # print("enter", type(self).__name__, id(self))
        self.root.frames.append(self)
        return self

    def __exit__(self, ex_type, value, traceback):
        tls.puistack.pop()
        # print("exit", type(self).__name__, id(self))
        self.root.frames.pop()
        if ex_type is None: # don't consume exception
            return self

    def content(self):
        return None

    def dump(self):
        if DEBUG:
            print(f"content() start", self.key)
        start = time.time()
        with self as scope:
            self.content()
        if DEBUG:
            print(f"content() time: {time.time()-start:.5f}", self.key)
        return scope

    def destroy(self, direct):
        try:
            PUIView.__ALLVIEWS__.remove(self)
        except:
            pass
        return super().destroy(direct)

    def redraw(self):
        if self.retired_by or self.destroyed:
            return
        self.dirty = True
        if self.updating:
            return
        self.updating = True
        self.sync()

    def wait(self):
        """
        Wait for the view to be updated.
        """
        while self.updating or self.dirty:
            time.sleep(0.0001)

    # Subview update entry point
    def sync(self):
        if not self.pui_virtual:
            raise VDomError(f"sync() called on non-virtual node {self.key}")
        if self.retired_by or self.destroyed:
            return
        self.dirty = False

        dom_parent = (self.pui_dom_parent or self).get_node()
        if DEBUG:
            print(f"Sync subview {self.key}@{id(self)} retired_by={id(self.retired_by) if self.retired_by else None} destroyed={self.destroyed} dom_parent={dom_parent.key}@{id(dom_parent)}")
        found, offset = dom_parent.findDomOffsetForNode(self)
        if DEBUG:
            print(f"    found={found} offset={offset}")
        if not found:
            if DEBUG:
                print(dom_parent.serialize(show_pyid=True, show_hierarchy=True))
            offset = 0

        last_children = self.children
        try:
            self.update()
        except (StateMutationInViewBuilderError, VDomError):
            raise
        except:
            # prevent crash in hot-reloading
            self.children = last_children
            import traceback
            print("## <ERROR OF content() >", self.key, id(self))
            traceback.print_exc()
            print("## </ERROR OF content()>")

        start = time.time()
        if DEBUG:
            print("sync() start", self.key)

        if DEBUG:
            print(f"offset for {self.key}@{id(self)} on {dom_parent.key}@{id(dom_parent)} is {offset}")
        sync(self, dom_parent, offset, last_children, self.children)
        if DEBUG:
            print(f"sync() time: {time.time()-start:.5f}", self.key)

        self.updating = False
        if self.dirty:
            self.redraw()

    def update(self, prev=None):
        if self.retired_by:
            return
        if self.destroyed:
            return
        if not prev:
            if self.setup:
                self.setup()
                self.setup = None
        else:
            self.setup = None

        if DEBUG:
            print(f"update() {self.key}@{id(self)} prev={id(prev) if prev else None}")

        if prev and prev is not self:
            prev.retired_by = self
            try:
                print(f"retired {prev.key}@{id(prev)}")
                PUIView.__ALLVIEWS__.remove(prev)
            except:
                pass

        self.children = []
        if DEBUG:
            print(f"content() start", self.key)
        start = time.time()
        with self as scope: # init frame stack
            self.content() # V-DOM builder
        if DEBUG:
            print(f"content() time: {time.time()-start:.5f}", self.key)

    def setup(self):
        pass

    def run(self):
        self.sync()
        self.start()



from .. import *
from textual import widgets, containers

class TBase(PUINode):
    scroll = False
    container_x = False # axis
    container_y = False # axis
    expand_x_prio = 0
    expand_y_prio = 0
    expand_x1_children = 0
    expand_x2_children = 0
    expand_x3_children = 0
    expand_x4_children = 0
    expand_y1_children = 0
    expand_y2_children = 0
    expand_y3_children = 0
    expand_y4_children = 0
    cached_tparent = None

    @property
    def expand_x(self):
        parent = self.cached_tparent
        expand = self.expand_x_prio
        if not parent:
            return False

        # textual handles 1fr as shrinkable, but we need scrolller's content not to shrink
        # See Exp.1 in refs/textual_layout.py
        if parent.scroll:
            return False

        # textual populates auto(1fr) to be 1fr(1fr), but we require expanding not to go over the container
        # See Exp.2 in refs/textual_layout.py
        if not parent.expand_x and expand < 3:
            return False

        return expand

    @property
    def expand_y(self):
        parent = self.cached_tparent
        expand = self.expand_y_prio
        if not parent:
            return False

        # textual handles 1fr as shrinkable, but we need scrolller's content not to shrink
        # See Exp.1 in refs/textual_layout.py
        if parent.scroll:
            return False

        # textual populates auto(1fr) to be 1fr(1fr), but we require expanding not to go over the container
        # See Exp.2 in refs/textual_layout.py
        if not parent.expand_y and expand < 3:
            return False

        return expand

    def tremove(self):
        self.ui.remove()

    def destroy(self, direct):
        self.ui.remove()
        return super().destroy(direct)

    def update(self, prev):
        super().update(prev)

        self.cached_tparent = parent = self.tparent
        if parent:
            if self.layout_weight:
                if parent.container_x:
                    self.expand_x_prio = 4
                if parent.container_y:
                    self.expand_y_prio = 4

            if self.expand_x_prio >= 1:
                parent.expand_x1_children += 1
            if self.expand_x_prio >= 2:
                parent.expand_x2_children += 1
            if self.expand_x_prio >= 3:
                parent.expand_x3_children += 1
            if self.expand_x_prio >= 4:
                parent.expand_x4_children += 1

            if self.expand_y_prio >= 1:
                parent.expand_y1_children += 1
            if self.expand_y_prio >= 2:
                parent.expand_y2_children += 1
            if self.expand_y_prio >= 3:
                parent.expand_y3_children += 1
            if self.expand_y_prio >= 4:
                parent.expand_y4_children += 1

    def postUpdate(self):
        super().postUpdate()
        parent = self.cached_tparent
        if parent:
            if parent.container_x:
                if self.expand_x_prio < 1 and parent.expand_x1_children > 0:
                    self.expand_x_prio = 0
                if self.expand_x_prio < 2 and parent.expand_x2_children > 0:
                    self.expand_x_prio = 0
                if self.expand_x_prio < 3 and parent.expand_x3_children > 0:
                    self.expand_x_prio = 0
                if self.expand_x_prio < 4 and parent.expand_x4_children > 0:
                    self.expand_x_prio = 0

            if parent.container_y:
                if self.expand_y_prio < 1 and parent.expand_y1_children > 0:
                    self.expand_y_prio = 0
                if self.expand_y_prio < 2 and parent.expand_y2_children > 0:
                    self.expand_y_prio = 0
                if self.expand_y_prio < 3 and parent.expand_y3_children > 0:
                    self.expand_y_prio = 0
                if self.expand_y_prio < 4 and parent.expand_y4_children > 0:
                    self.expand_y_prio = 0

        self.t_update_layout()

    @property
    def tparent(self):
        parent = self.parent
        while not isinstance(parent, TBase):
            if parent==parent.parent:
                parent = None
                break
            parent = parent.parent
        return parent

    def t_update_layout(self):
        if not self.ui:
            return

        width = "auto"
        if self.expand_x:
            width = "1fr"

        height = "auto"
        if self.expand_y:
            height = "1fr"

        if self._debug:
            print("layout", self.key, f"{width}:{height} expand_x={self.expand_x}", f"expand_y={self.expand_y}", f"expand_x_prio={self.expand_x_prio}", f"expand_y_prio={self.expand_y_prio}")
        self.ui.styles.width = width
        self.ui.styles.height = height

class TPUIView(PUIView):
    pui_virtual = True



from .. import *
from .base import *
from .label import ClickableQLabel
from PySide6.QtWidgets import QSizePolicy
import os

class Image(QtBaseWidget):
    def __init__(self, path):
        super().__init__()
        self.path = path
        self.pixmap = None

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.eventFilter = prev.eventFilter
            self.curr_path = prev.curr_path
            self.curr_path_mtime = prev.curr_path_mtime
            self.pixmap = prev.pixmap
        else:
            self.ui = ClickableQLabel()
            self.ui.clicked.connect(self._clicked)
            self.curr_path = Prop()
            self.curr_path_mtime = Prop()

        if self._onClicked:
            self.ui.setCursor(QtCore.Qt.PointingHandCursor)

        if self.layout_weight:
            # XXX keep aspect ratio
            self.ui.setScaledContents(True)
            self.ui.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Ignored))
        else:
            self.ui.setScaledContents(False)
            self.ui.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred))

        if self.curr_path.set(self.path) or self.curr_path_mtime.set(os.path.getmtime(self.path)):
            self.pixmap = QtGui.QPixmap(self.path)
            if self.layout_width is not None and self.layout_height is not None:
                self.pixmap = self.pixmap.scaled(self.layout_width, self.layout_height, QtCore.Qt.KeepAspectRatio, mode=QtCore.Qt.SmoothTransformation)
            elif self.layout_width is not None:
                self.pixmap = self.pixmap.scaledToWidth(self.layout_width, mode=QtCore.Qt.SmoothTransformation)
            elif self.layout_height is not None:
                self.pixmap = self.pixmap.scaledToHeight(self.layout_height, mode=QtCore.Qt.SmoothTransformation)

            self.ui.setPixmap(self.pixmap)

        super().update(prev)



from .. import *
from .base import *

class Divider(FBase):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            try:
                self.ui.update()
            except:
                pass
        else:
            self.ui = ft.Divider()
        self.ui.expand = self.layout_weight



from .. import *
from .base import *

class QtApplicationSignal(QtCore.QObject):
    quit = QtCore.Signal()

class Application(QtPUIView):
    def __init__(self, icon=None):
        super().__init__()
        self.ui = None
        self.icon = icon
        self._qtappsignal = QtApplicationSignal()
        self._qtappsignal.quit.connect(self._quit, QtCore.Qt.ConnectionType.QueuedConnection) # Use QueuedConnection to prevent nested trigger

    def redraw(self):
        if self.ui:
            super().redraw()
        else:
            self.sync()

    def update(self, prev=None):
        if not self.ui:
            from PySide6 import QtWidgets
            self.ui = QtWidgets.QApplication([])
            if self.icon:
                self.ui.setWindowIcon(QtGui.QIcon(self.icon))

        super().update(prev)

    def addChild(self, idx, child):
        child.outer.show()

    def removeChild(self, idx, child):
        child.outer.close()

    def start(self):
        self.ui.exec()

    def quit(self):
        self._qtappsignal.quit.emit()

    def _quit(self):
        self.ui.quit()

def PUIApp(func):
    def func_wrapper(*args, **kwargs):
        class PUIAppWrapper(Application):
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args, **kwargs)

        ret = PUIAppWrapper(func.__name__)
        return ret

    return func_wrapper



from .. import *
from .base import *
import math

class Scroll(TBase):
    END = -0.0
    scroll = True
    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        self.align_x = 0
        self.align_y = 0
        super().__init__()

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = containers.ScrollableContainer()

        v = "auto"
        self.container_y = True
        self.expand_y_prio = 3
        if self.vertical is True:
            v = "scroll"
        elif self.vertical is False:
            v = "hidden"
            self.container_y = False
            self.expand_y_prio = 1

        h = "auto"
        self.container_x = True
        self.expand_x_prio = 3
        if self.horizontal is True:
            h = "scroll"
        elif self.horizontal is False:
            h = "hidden"
            self.container_x = False
            self.expand_x_prio = 1

        self.ui.set_styles(f"overflow-x: {h}; overflow-y: {v};")

        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, TBase):
            self.inner.mount(child.outer)
            self.t_update_layout()
        else:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, TBase):
            child.tremove()
        else:
            self.removeChild(idx, child.children[0])

    def scrollX(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_x = 0
            self.hsb_offset = pos
        else:
            self.align_x = 1
            self.hsb_offset = abs(pos)
        return self

    def scrollY(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_y = 0
            self.vsb_offset = pos
        else:
            self.align_y = 1
            self.vsb_offset = abs(pos)
        return self



from .. import *
from .base import *

class Window(WxBaseWidget):
    pui_terminal = False

    def __init__(self, title=None, icon=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.title = title
        self.icon = icon
        self.size = size
        self.curr_icon = None
        self.curr_size = None
        self.maximize = maximize
        self.curr_maximize = None
        self.fullscreen = fullscreen
        self.curr_fullscreen = None

    @property
    def inner(self):
        return self.panel

    def update(self, prev=None):
        if prev and prev.ui:
            self.ui = prev.ui
            self.panel = prev.panel
            self.curr_icon = prev.curr_icon
            self.curr_size = prev.curr_size
            self.curr_maximize = prev.curr_maximize
            self.curr_fullscreen = prev.curr_fullscreen
        else:
            self.ui = wx.Frame(None)
            self.panel = wx.Panel(self.ui) # to grab focus when clicked
            self.curr_icon = Prop()
            self.curr_size = Prop()
            self.curr_maximize = Prop()
            self.curr_fullscreen = Prop()

        if self.curr_icon.set(self.icon):
            self.ui.SetIcon(wx.Icon(self.icon))
        if self.curr_size.set(self.size):
            self.ui.SetSize(wx.Size(*self.size))
        if self.curr_maximize.set(self.maximize):
            self.ui.Maximize(True)
        if self.curr_fullscreen.set(self.fullscreen):
            self.ui.ShowFullScreen(True)
        if not self.title is None:
            self.ui.SetTitle(self.title)
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, WxBaseLayout):
            self.ui.SetSizer(child.outer)
        elif isinstance(child, WxBaseWidget):
            pass

    def removeChild(self, idx, child):
        if isinstance(child, WxBaseLayout):
            pass
        elif isinstance(child, WxBaseWidget):
            pass

    def postSync(self):
        self.ui.Layout()



from .. import *
from .base import *

class Checkbox(TkBaseWidget):
    def __init__(self, text, model, value=None):
        super().__init__()
        self.text = text
        self.model = model
        self.value = text if value is None else value

    def update(self, prev):
        if prev and prev.ui:
            self.var = prev.var
            self.ui = prev.ui
        else:
            self.var = tk.IntVar()
            self.ui = tk.Checkbutton(self.tkparent.inner, text=self.text, variable=self.var, anchor="w")
        self.ui.configure(command=self._select)
        if checkbox_get(self.model, self.value):
            self.ui.select()
        else:
            self.ui.deselect()

        super().update(prev)

    def _select(self):
        node = self.get_node()
        checkbox_set(self.model, node.var.get() != 0, self.value)



from .. import *
from .base import *

class Checkbox(WxBaseWidget):
    def __init__(self, text, model, value=None):
        super().__init__()
        self.text = text
        self.model = model
        self.value = text if value is None else value

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.SetLabel(self.text)
        else:
            self.ui = wx.CheckBox(getWindow(self.parent), label=self.text)
            self.ui.Bind(wx.EVT_CHECKBOX, self._checked)

        self.ui.SetValue(checkbox_get(self.model, self.value))

        super().update(prev)

    def _checked(self, *args, **kwargs):
        node = self.get_node()
        checkbox_set(node.model, node.ui.GetValue(), self.value)
        node._clicked()


from threading import Timer
from .view import *

class TimelineView(PUINode):
    pui_virtual = True
    def __init__(self, ttl_sec):
        super().__init__()
        self.timer = None
        self.ttl_sec = ttl_sec

    def update(self, prev):
        if prev and hasattr(prev, "timer"):
            self.timer = prev.timer
        else:
            self.timer = Timer(self.ttl_sec, self.timer_cb)
            self.timer.setDaemon(True)
            self.timer.start()

    def timer_cb(self):
        if not self.timer:
            return
        node = self.get_node()
        root = node.root
        if not root:
            return
        root.redraw()
        node.timer = Timer(self.ttl_sec, self.timer_cb)
        node.timer.setDaemon(True)
        node.timer.start()

    def destroy(self, direct):
        timer = self.timer
        self.timer = None
        if timer:
            timer.cancel()
        super().destroy(direct)



from PySide6.QtWidgets import QSizePolicy, QLayout

from .application import *
from .button import *
from .canvas import *
from .checkbox import *
from .combobox import *
from .dialog import *
from .divider import *
from .image import *
from .label import *
from .layout import *
from .modal import *
from .matplotlib import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .splitter import *
from .table import *
from .tab import *
from .text import *
from .textfield import *
from .tree import *
from .window import *
from .mdi import *
from .toolbar import *

PUIView = QtPUIView

def PUI(func):
    """
    PUI.PySide6.PUI triggers update() by signal/slot
    """
    def func_wrapper(*args, **kwargs):
        class PUIViewWrapper(QtPUIView):
            pui_virtual = True
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args, **kwargs)

        ret = PUIViewWrapper(func.__name__)
        return ret

    return func_wrapper

PUI_BACKEND = "PySide6"



from .application import *
from .button import *
from .canvas import *
from .checkbox import *
from .label import *
from .layout import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .tab import *
from .text import *
from .textfield import *
from .window import *

PUIView = TPUIView

def PUI(func):
    """
    PUI.tkinter.PUI triggers update() by .after()
    """
    def func_wrapper(*args, **kwargs):
        class PUIViewWrapper(TPUIView):
            pui_virtual = True
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args, **kwargs)

        ret = PUIViewWrapper(func.__name__)
        return ret

    return func_wrapper
class DummyWidget(PUINode):
    pui_supported = False
    def __init__(self, *args, **kwargs):
        super().__init__()

Combobox = DummyWidget
ComboboxItem = DummyWidget
Divider = DummyWidget

PUI_BACKEND = "tkinter"
